@@ -2,7 +2,8 @@
 DIR=web/xhotkeys
 FILES="index-template.html
 howtouse-template.html
-contact-template.html"
+contact-template.html
+tokland.css"
 EDITOR=bluefish
 
 (cd $DIR; $EDITOR $FILES) &
