@@ -22,8 +22,8 @@
 # Feel free to contact me at <arnau@ehas.org>
 
 # Python modules
-import os, sys, time, signal, urllib
-import re, optparse, string, select
+import os, sys, time, signal
+import urllib, re, optparse
 import inspect, threading
 
 # Xlib modules
@@ -47,6 +47,9 @@ NAME = "xhotkeys"
 DISABLE_HOTKEY = Xlib.XK.XK_BackSpace
 DISABLED_STRING = "Disabled"
 MOUSE_STRING = "Button"
+SETTINGS_WINDOW = "window: %d,%d (%dx%d)"
+SETTINGS_COLUMN = "table: %d,%d,%d"
+TARGET_TYPE_URI_LIST = 80
 
 ## Verbose levels
 LEVELS_STRING = "error", "info", "debug"
@@ -189,7 +192,7 @@ class XhotkeysServer(threading.Thread):
 
 	#######################################
 	def ungrab(self):
-		"""Ungrab keycoard bindings"""
+		"""Ungrab keyboard and mouse bindings"""
 		self.display.ungrab_keyboard(Xlib.X.CurrentTime)
 		self.display.flush()
 		self.root.ungrab_key(Xlib.X.AnyKey, Xlib.X.AnyModifier)
@@ -272,8 +275,8 @@ class XhotkeysServer(threading.Thread):
 						if comb["type"] == "button" and comb["keycode"] == button and comb["mask"] == mask:
 							run_command(comb["command"])
 			
-			# Give the CPU a breath
-			time.sleep(0.2)
+			# Give the CPU a breath (1/4 second should be fine)
+			time.sleep(0.25)
 		
 		self.ungrab()
 
@@ -281,10 +284,11 @@ class XhotkeysServer(threading.Thread):
 #######################
 class Xhotkeys:
 	####################################
-	def __init__(self, cfile, pidfile, verbose_level=DEBUG):
+	def __init__(self, cfile, pidfile, settingsfile, verbose_level=DEBUG):
 		self.cfile = cfile
 		self.verbose_level = verbose_level 
 		self.pidfile = pidfile
+		self.settingsfile = settingsfile
 		self.display = Xlib.display.Display()
 		self.root = self.display.screen().root
 		self.hkserver = None
@@ -476,7 +480,7 @@ class Xhotkeys:
 		newcomb = create_dict(mask=mask, keycode=keycode, hotkey=hotkey)
 
 		for index, comb in enumerate(self.combinations):
-			if index == self.active_row or not comb["hotkey"]: continue
+			if comb == self.old_comb: continue
 			if comb["hotkey"] == newcomb["hotkey"]:
 				text =  "This hotkey (%s) is already used by %s" %(newcomb["hotkey"], comb["name"])
 				self.error_message(text)
@@ -493,7 +497,6 @@ class Xhotkeys:
 
 	#######################################
 	def open_add_window(self):
-		self.old_comb = self.current_comb.copy()
 		self.widgets["mainwindow"].set_sensitive(False)
 		self.widgets["name"].set_text(self.current_comb["name"])
 		self.widgets["command"].set_text(self.current_comb["command"])
@@ -538,6 +541,7 @@ class Xhotkeys:
 
 	#######################################
 	def exit_conf(self):
+		self.save_settings()
 		gtk.main_quit()
 		if self.hkserver and self.hkserver.isAlive():
 			self.stop_server()
@@ -555,18 +559,18 @@ class Xhotkeys:
 			getattr(self.widgets[key], function)(value)
 
 	#########################################
-	def get_current_row(self):
-		liststore, rows = self.listview.get_selection().get_selected_rows()
-		if not rows or len(rows) < 1 or len(rows[0]) < 1: return -1
-		return rows[0][0]
+	def get_current_rows(self):
+		return self.treeselection.get_selected_rows()[1]
 	
 	#######################################
-	def delete_entry(self, widget, answer, iter, row):
+	def delete_entry(self, widget, answer, iters, rows):
 		widget.destroy()
 		if answer == gtk.RESPONSE_YES:
-			self.liststore.remove(iter)
-			del self.combinations[row]
-			self.active_row = -1
+			for iter in iters:
+				self.liststore.remove(iter)
+			for index, entry in enumerate(self.combinations[:]):
+				if index in rows:
+					self.combinations.remove(entry)
 			self.apply_widgets("set_sensitive", delete=False, modify=False)
 			self.save_and_reload()
 		self.widgets["mainwindow"].set_sensitive(True)
@@ -606,9 +610,9 @@ class Xhotkeys:
 	#######################################
 
 	#######################################
-	def on_mainwindow_destroy(self, widget):
+	def on_mainwindow_delete_event(self, widget, event):
 		self.exit_conf()
-
+		
 	#######################################
 	def on_addwindow_delete_event(self, widget, event):
 		self.on_cancel_clicked(widget)
@@ -626,12 +630,13 @@ class Xhotkeys:
 		name, command = ret
 		hotkey = self.keycodes_to_string([])
 		self.current_comb = create_dict(name=name, command=command, hotkey=hotkey)
+		self.old_comb = None
 		self.open_add_window()
+		self.old_comb = None
 		self.textbox_check(self.widgets["name"], "name")
 		self.textbox_check(self.widgets["command"], "command")
 		self.new_window = "add"
 
-
 	#######################################
 	def on_accept_clicked(self, widget):
 		for index, combination in enumerate(self.combinations):
@@ -642,15 +647,16 @@ class Xhotkeys:
 
 		if self.new_window == "modify":
 			self.combinations[self.active_row] = self.current_comb
-			liststore, rows = self.listview.get_selection().get_selected_rows()			
-			model, ite = self.listview.get_selection().get_selected()
-			liststore.set_value(ite, 0, self.current_comb["name"])
-			liststore.set_value(ite, 1, self.current_comb["command"])
-			liststore.set_value(ite, 2, self.current_comb["hotkey"].title())
+			rows = self.get_current_rows()
+			treemodel, selected = self.treeselection.get_selected_rows()
+			iter = [treemodel.get_iter(path) for path in selected][0]
+			treemodel.set_value(iter, 0, self.current_comb["name"])
+			treemodel.set_value(iter, 1, self.current_comb["command"])
+			treemodel.set_value(iter, 2, self.current_comb["hotkey"].title())
 		else:
 			self.combinations.append(self.current_comb)
 			self.liststore.insert(len(self.combinations)-1, [self.current_comb["name"], self.current_comb["command"], self.current_comb["hotkey"].title()])
-			self.listview.set_cursor(len(self.combinations)-1)
+			self.treeview.set_cursor(len(self.combinations)-1)
 
 		self.widgets["addwindow"].hide()
 		self.widgets["mainwindow"].set_sensitive(True)
@@ -662,10 +668,16 @@ class Xhotkeys:
 	def on_cancel_clicked(self, widget):
 		self.widgets["addwindow"].hide()
 		self.widgets["mainwindow"].set_sensitive(True)
-		self.current_comb = self.get_combinations(self.active_row)
+		if self.old_comb:
+			self.current_comb = self.old_comb.copy()
 
 	#######################################
 	def on_modify_clicked(self, widget):
+		row = self.get_current_rows()[0][0]
+		self.active_row = row
+		self.current_comb = self.combinations[row].copy()
+		self.old_comb = self.current_comb.copy()
+		treemodel, selected = self.treeselection.get_selected_rows()
 		self.open_add_window()
 		self.new_window = "modify"
 
@@ -680,6 +692,7 @@ class Xhotkeys:
 		self.current_comb = create_dict(name="", command="", hotkey=hotkey)
 		self.apply_widgets("set_sensitive", accept=False, test=False)		
 		self.widgets["name"].grab_focus()
+		self.old_comb = None
 		self.open_add_window()
 		self.new_window = "add"
 		
@@ -744,15 +757,19 @@ class Xhotkeys:
 
 	#######################################
 	def on_delete_clicked(self, widget):
-		model, iter = self.listview.get_selection().get_selected()
+		rows = self.get_current_rows()
+		model, selected = self.treeselection.get_selected_rows()
+		iters = [model.get_iter(path) for path in selected]
+		names = [self.combinations[row[0]]["name"] for row in rows]
+		todelete = "\n".join(names)
 		message = gtk.MessageDialog(self.widgets["mainwindow"], gtk.DIALOG_MODAL, \
-			gtk.MESSAGE_ERROR, gtk.BUTTONS_YES_NO, "Are you sure you want to delete entry '%s'?" \
-			%self.current_comb["name"])
+			gtk.MESSAGE_WARNING, gtk.BUTTONS_YES_NO, "Are you sure you want to delete the following entries?\n%s" \
+			%todelete)
 		message.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
-		message.connect("response", self.delete_entry, iter, self.active_row)
+		rowstodelete = [row[0] for row in rows]
+		message.connect("response", self.delete_entry, iters, rowstodelete)
 		self.widgets["mainwindow"].set_sensitive(False)
 		message.show()
-		return
 
 	#######################################
 	def on_browse_clicked(self, widget):
@@ -768,37 +785,37 @@ class Xhotkeys:
 	def on_command_changed(self, widget):
 		self.textbox_check(widget, "command")
 		self.widgets["test"].set_sensitive(self.widgets["command"].get_text() != "")
-			
 
 	#######################################
 	def on_hotkeytable_row_activated(self, widget, rows, column):
-		self.open_add_window()
-		self.new_window = "modify"
+		self.on_modify_clicked(self.widgets["modify"])
 
 	#######################################
 	def on_hotkeytable_button_press_event(self, treeview, event):
-		if event.button != 3: return
-		menu = gtk.Menu()
 		if not treeview.get_path_at_pos(int(event.x), int(event.y)): 
-			items = ("gtk-add", self.on_add_clicked),
-		else: items = ("gtk-edit", self.on_modify_clicked), ("gtk-delete", self.on_delete_clicked)
-		for stock_id, connect in items:
-			item = gtk.ImageMenuItem(stock_id=stock_id)
-			menu.append(item)
-			item.connect("activate", connect)
-			item.show()
-		menu.popup( None, None, None, 3, event.time)
+			if event.button != 3: self.treeselection.unselect_all()
+			else: items = ("gtk-add", self.on_add_clicked),
+		elif event.button == 3: 
+			self.rbutton_event = event.time
 	
 	#########################################
-	def on_hotkeytable_cursor_changed(self, *args):
-		row = self.get_current_row()
-		if row < 0: 
-			self.apply_widgets("set_sensitive", delete=False, modify=False)
-			self.active_row = -1
-		else:
-			self.apply_widgets("set_sensitive", delete=True, modify=True)
-			self.current_comb = self.get_combinations(row)
-			self.active_row = row
+	def hotkeytable_selection_changed(self, *args):
+		rows = self.get_current_rows()
+		if self.rbutton_event:
+			menu = gtk.Menu()
+			items = [("gtk-delete", self.on_delete_clicked)]
+			if len(rows) == 1:
+				items.append(("gtk-edit", self.on_modify_clicked))
+			for stock_id, connect in items:
+				item = gtk.ImageMenuItem(stock_id=stock_id)
+				menu.append(item)
+				item.connect("activate", connect)
+				item.show()
+			menu.popup(None, None, None, 3, self.rbutton_event)
+			self.rbutton_event = None
+		if len(rows) == 0: self.apply_widgets("set_sensitive", delete=False, modify=False)
+		elif len(rows) == 1: self.apply_widgets("set_sensitive", delete=True, modify=True)
+		else: self.apply_widgets("set_sensitive", delete=True, modify=False)
 	
 	#####################################################
 	#####################################################
@@ -847,10 +864,14 @@ class Xhotkeys:
 		
 	#####################################################
 	def create_table(self, widget_name, options):
-		self.listview = self.tree.get_widget("hotkeytable")
+		self.treeview = self.tree.get_widget(widget_name)
+		self.treeselection = self.treeview.get_selection()
+		self.rbutton_event =None
+		self.treeselection.connect("changed", self.hotkeytable_selection_changed)
+		self.treeselection.set_mode(gtk.SELECTION_MULTIPLE)
 		format = tuple([str]) * len(options)
 		self.liststore = gtk.ListStore(*format)
-		self.listview.set_model(self.liststore)			
+		self.treeview.set_model(self.liststore)			
 		for index, (name, min_width, expand) in enumerate(options):
 			renderer = gtk.CellRendererText()
 			column = gtk.TreeViewColumn(name, renderer, text=index)
@@ -859,15 +880,47 @@ class Xhotkeys:
 			column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
 			column.set_expand(expand)
 			column.set_min_width(min_width)
-			self.listview.append_column(column)
+			self.treeview.append_column(column)
 		self.apply_widgets("set_sensitive", delete=False, modify=False)
-		TARGET_TYPE_URI_LIST = 80
 		list = [ ( 'text/uri-list', 0, TARGET_TYPE_URI_LIST ) ]
-		self.listview.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT | \
+		self.treeview.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT | \
 			gtk.DEST_DEFAULT_DROP, list, gtk.gdk.ACTION_COPY)
 
-		self.active_row = -1
 
+	#####################################################
+	def save_settings(self):
+		try: fd = open(self.settingsfile, "w")
+		except: self.debug("settings could not be opened for writing: %s" %self.settingsfile); return
+		x, y = self.widgets["mainwindow"].get_position()
+		width, height = self.widgets["mainwindow"].get_size()
+		fd.write(SETTINGS_WINDOW % (x, y, width, height) + "\n")
+		widths = tuple([column.get_width() for column in self.widgets["hotkeytable"].get_columns()])
+		fd.write(SETTINGS_COLUMN % widths + "\n")
+		fd.close()
+		self.debug("window settings saved")
+				
+	#####################################################
+	def load_settings(self):
+		try: fd = open(self.settingsfile)
+		except: self.debug("settings file not found: %s" %self.settingsfile); return
+		match_window = SETTINGS_WINDOW.replace(" ", "\s+").replace("(", "\(").replace(")", "\)").replace("%d", "(\d+)")
+		match_column = SETTINGS_COLUMN.replace(" ", "\s+").replace("%d", "(\d+)")
+		for line in fd.readlines():
+			match = re.findall(match_window, line)
+			if match:
+				try: x, y, width, height = [int(x) for x in match[0]]
+				except: continue
+				self.widgets["mainwindow"].move(x, y)
+				self.widgets["mainwindow"].resize(width, height)
+			match = re.findall(match_column, line)
+			if match:
+				try: widths = [int(x) for x in match[0]]
+				except: continue
+				columns = self.widgets["hotkeytable"].get_columns()
+				[columns[x].set_fixed_width(widths[x]) for x in range(len(columns))]
+		fd.close()
+		self.debug("window settings loaded")
+			
 	#####################################################
 	def config(self):
 		self.state = "config"
@@ -878,21 +931,15 @@ class Xhotkeys:
 		table = ("Name", 100, False), ("Command", 300, True), ("Hotkey", 150, True)
 		self.read_configuration(self.cfile)
 		self.create_table("hotkeytable", table)		
+		self.load_settings()
 		for comb in self.combinations:
 			self.liststore.append([comb["name"], comb["command"], comb["hotkey"].title()])
-		
 		if self.check_active():
 			self.widgets["status"].push(0, " Xhotkeys daemon is running (pid %d)" %self.daemon_pid)
-		else:
-			# There is no hotkey server running, run it temporally
-			self.widgets["status"].push(0, " Xhotkeys daemon is not running")
-			self.debug("starting Xhotkeys server")
-			self.hkserver = XhotkeysServer(self.verbose_level)
-			self.hkserver.set_combination(self.combinations)
-			self.hkserver.start()
-			gobject.idle_add(self.end_combinations_keys)
-			
+		else: self.widgets["status"].push(0, " Xhotkeys daemon is not running")
+		
 		self.configure_signals([signal.SIGTERM, signal.SIGINT])
+		self.widgets["mainwindow"].show()
 		gtk.main()
 
 	#########################################
@@ -930,14 +977,15 @@ def main():
 Bind keys and mouse combinations to commands in the X-Window System""" 
 	home, user = os.getenv("HOME"), os.getenv("USER")
 	default_cfile = os.path.join(home, ".%s" %NAME)
-	default_pidfile = os.path.join(home, ".%s.%s" %(NAME, user))
+	default_pidfile = os.path.join(home, ".%s.pid-%s" %(NAME, user))
+	default_settings = os.path.join(home, ".%s.settings" %NAME)
 	parser = optparse.OptionParser(usage)
 	parser.add_option('-v', '--verbose', default=0, dest='verbose_level', action="count", help='Increase verbose level')
 	parser.add_option('-c', '--config', dest='config', default=False, action='store_true', help='Open a GTK+ interface configurator')
 	parser.add_option('-f', '--configuration-file', dest='cfile', default=default_cfile, metavar='FILE', type='string', help='Alternative configuration file')
-	parser.add_option('-p', '--pidfile', dest='pidfile', default=default_pidfile, metavar='FILE', type='string', help='Alternative pidfile')
 	options, args = parser.parse_args()
-	xhk = Xhotkeys(options.cfile, options.pidfile, options.verbose_level)
+	
+	xhk = Xhotkeys(options.cfile, default_pidfile, default_settings, verbose_level=options.verbose_level)
 	if options.config: xhk.config()
 	else: xhk.server()
 	
